from itertools import chain
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Set,
    Type,
    Union,
    cast,
    no_type_check,
)

from django.db.models import Field, ManyToManyRel, ManyToOneRel
from pydantic.fields import FieldInfo

from ..errors import ConfigError
from ..pydanticutils import IS_PYDANTIC_V1, compute_field_annotations
from .getters import DjangoGetter
from .mixins import SchemaMixins
from .model_validators import ModelValidatorGroup
from .schema_registry import registry as global_registry
from .utils.converter import convert_django_field_with_choices

ALL_FIELDS = "__all__"

__all__ = ["ModelSchema"]

if IS_PYDANTIC_V1:
    from pydantic import BaseConfig, BaseModel, PyObject
    from pydantic.class_validators import (
        extract_root_validators,
        extract_validators,
        inherit_validators,
    )
    from pydantic.fields import ModelField, Undefined
    from pydantic.main import (
        ANNOTATED_FIELD_UNTOUCHED_TYPES,
        UNTOUCHED_TYPES,
        ModelMetaclass,
        generate_hash_function,
        validate_custom_root_type,
    )
    from pydantic.typing import get_args, get_origin, is_classvar, resolve_annotations
    from pydantic.utils import (
        ROOT_KEY,
        ClassAttribute,
        generate_model_signature,
        is_valid_field,
        lenient_issubclass,
        unique_list,
        validate_field_name,
    )

    namespace_keys = [
        "__config__",
        "__fields__",
        "__validators__",
        "__pre_root_validators__",
        "__post_root_validators__",
        "__schema_cache__",
        "__json_encoder__",
        "__custom_root_type__",
        "__private_attributes__",
        "__slots__",
        "__hash__",
        "__class_vars__",
        "__annotations__",
    ]

    class PydanticNamespace:
        __annotations__: Dict = {}
        __config__: Optional[Type[BaseConfig]] = None
        __fields__: Dict[str, ModelField] = {}
        __validators__: ModelValidatorGroup = ModelValidatorGroup({})
        __pre_root_validators__: List = []
        __post_root_validators__: List = []
        __custom_root_type__ = None
        __private_attributes__ = None
        __class_vars__: Set = set()

        def __init__(self, cls: Type):
            for key in namespace_keys:
                value = getattr(cls, key, getattr(self, key, None))
                setattr(self, key, value)

    def update_class_missing_fields(
        cls: Type, bases: List[Type[BaseModel]], namespace: Dict
    ) -> Type[BaseModel]:
        old_namespace: PydanticNamespace = PydanticNamespace(cls)
        fields = old_namespace.__fields__ or {}
        config: Type[BaseConfig] = cast(Type[BaseConfig], old_namespace.__config__)
        validators: "ValidatorListDict" = {}

        pre_root_validators, post_root_validators = [], []
        class_vars: Set[str] = set()
        hash_func: Optional[Callable[[Any], int]] = None
        untouched_types = ANNOTATED_FIELD_UNTOUCHED_TYPES

        def is_untouched(val: Any) -> bool:
            return (
                isinstance(val, untouched_types)
                or val.__class__.__name__ == "cython_function_or_method"
            )

        for base in reversed(bases):
            if base != BaseModel:
                validators = inherit_validators(base.__validators__, validators)
                pre_root_validators += base.__pre_root_validators__
                post_root_validators += base.__post_root_validators__
                class_vars.update(base.__class_vars__)
                hash_func = base.__hash__

        validators = inherit_validators(extract_validators(namespace), validators)
        vg = ModelValidatorGroup(validators)
        new_annotations = resolve_annotations(
            namespace.get("__annotations__") or {}, getattr(cls, "__module__", None)
        )
        # annotation only fields need to come first in fields
        for ann_name, ann_type in new_annotations.items():
            if is_classvar(ann_type):
                class_vars.add(ann_name)
            elif is_valid_field(ann_name):
                validate_field_name(bases, ann_name)
                value = namespace.get(ann_name, Undefined)
                allowed_types = (
                    get_args(ann_type) if get_origin(ann_type) is Union else (ann_type,)
                )
                if (
                    is_untouched(value)
                    and ann_type != PyObject
                    and not any(
                        lenient_issubclass(get_origin(allowed_type), Type)  # type: ignore
                        for allowed_type in allowed_types
                    )
                ):
                    continue
                fields[ann_name] = ModelField.infer(
                    name=ann_name,
                    value=value,
                    annotation=ann_type,
                    class_validators=vg.get_validators(ann_name),
                    config=config,
                )

        untouched_types = UNTOUCHED_TYPES + config.keep_untouched
        for var_name, value in namespace.items():
            can_be_changed = var_name not in class_vars and not is_untouched(value)
            if (
                is_valid_field(var_name)
                and var_name not in new_annotations
                and can_be_changed
            ):
                validate_field_name(bases, var_name)
                inferred = ModelField.infer(
                    name=var_name,
                    value=value,
                    annotation=new_annotations.get(var_name, Undefined),
                    class_validators=vg.get_validators(var_name),
                    config=config,
                )
                if var_name in fields and inferred.type_ != fields[var_name].type_:
                    raise TypeError(
                        f"The type of {cls.__name__}.{var_name} differs from the new default value; "
                        f"if you wish to change the type of this field, please use a type annotation"
                    )
                fields[var_name] = inferred

        _custom_root_type = ROOT_KEY in fields
        if _custom_root_type:
            validate_custom_root_type(fields)
        vg.check_for_unused()

        old_namespace.__annotations__.update(new_annotations)
        pre_rv_new, post_rv_new = extract_root_validators(namespace)

        if hash_func is None:
            hash_func = generate_hash_function(config.frozen)
        pre_root_validators.extend(pre_rv_new)
        post_root_validators.extend(post_rv_new)

        new_namespace = {
            "__annotations__": old_namespace.__annotations__,
            "__config__": config,
            "__fields__": fields,
            "__validators__": vg.validators,
            "__pre_root_validators__": unique_list(pre_root_validators),
            "__post_root_validators__": unique_list(post_root_validators),
            "__class_vars__": class_vars,
            "__hash__": hash_func,
        }
        for k, v in new_namespace.items():
            if hasattr(cls, k):
                setattr(cls, k, v)
        # set __signature__ attr only for model class, but not for its instances
        cls.__signature__ = ClassAttribute(
            "__signature__", generate_model_signature(cls.__init__, fields, config)
        )
        return cls

    if TYPE_CHECKING:
        from pydantic.class_validators import ValidatorListDict
else:
    from pydantic import BaseConfig, BaseModel
    from pydantic._internal._model_construction import ModelMetaclass
    from pydantic.fields import Field as ModelField
    from pydantic_core import PydanticUndefined as Undefined

    PydanticNamespace = None

    def update_class_missing_fields(
        cls: Type, bases: List[Type[BaseModel]], namespace: Dict
    ):  # pragma: no cover
        return cls


class ModelSchemaConfigAdapter:
    def __init__(self, config: Dict) -> None:
        self.__dict__ = config


class ModelSchemaConfig:
    def __init__(
        self,
        schema_class_name: str,
        options: Optional[Union[Dict[str, Any], ModelSchemaConfigAdapter]] = None,
    ):
        self.model = getattr(options, "model", None)
        _include = getattr(options, "include", None) or ALL_FIELDS
        self.include = set() if _include == ALL_FIELDS else set(_include or ())
        self.exclude = set(getattr(options, "exclude", None) or ())
        self.skip_registry = getattr(options, "skip_registry", False)
        self.registry = getattr(options, "registry", global_registry)
        self.abstract = getattr(options, "ninja_schema_abstract", False)
        _optional = getattr(options, "optional", None)
        self.optional = (
            {ALL_FIELDS} if _optional == ALL_FIELDS else set(_optional or ())
        )
        self.depth = int(getattr(options, "depth", 0))
        self.schema_class_name = schema_class_name
        if not self.abstract:
            self.validate_configuration()
            self.process_build_schema_parameters()

    @classmethod
    def clone_field(cls, field: FieldInfo, **kwargs: Any) -> FieldInfo:
        field_dict = dict(field.__repr_args__())
        field_dict.update(**kwargs)
        new_field = FieldInfo(**field_dict)  # type: ignore
        return new_field

    def model_fields(self) -> Iterator[Field]:
        """returns iterator with all the fields that can be part of schema"""
        for fld in self.model._meta.get_fields():  # type: ignore
            if isinstance(fld, (ManyToOneRel, ManyToManyRel)):
                # skipping relations
                continue
            yield cast(Field, fld)

    def validate_configuration(self) -> None:
        self.include = set() if self.include == ALL_FIELDS else set(self.include or ())

        if not self.model:
            raise ConfigError("Invalid Configuration. 'model' is required")

        if self.include and self.exclude:
            raise ConfigError(
                "Only one of 'include' or 'exclude' should be set in configuration."
            )

    def check_invalid_keys(self, **field_names: Dict[str, Any]) -> None:
        keys = field_names.keys()
        invalid_include_exclude_fields = (
            set(self.include or []) | set(self.exclude or [])
        ) - keys
        if invalid_include_exclude_fields:
            raise ConfigError(
                f"Field(s) {invalid_include_exclude_fields} are not in model."
            )
        if ALL_FIELDS not in self.optional:
            invalid_options_fields = set(self.optional) - keys
            if invalid_options_fields:
                raise ConfigError(
                    f"Field(s) {invalid_options_fields} are not in model."
                )

    def is_field_in_optional(self, field_name: str) -> bool:
        if not self.optional:
            return False
        if ALL_FIELDS in self.optional:
            return True
        if (
            isinstance(self.optional, (set, tuple, list))
            and field_name in self.optional
        ):
            return True
        return False

    def process_build_schema_parameters(self) -> None:
        model_pk = getattr(
            self.model._meta.pk,  # type: ignore
            "name",
            self.model._meta.pk.attname,  # type: ignore
        )  # no type:ignore
        if (
            model_pk not in self.include
            and model_pk not in self.exclude
            and ALL_FIELDS not in self.optional
        ):
            self.optional.add(str(model_pk))


class ModelSchemaMetaclass(ModelMetaclass):
    @no_type_check
    def __new__(
        mcs,
        name: str,
        bases: tuple,
        namespace: dict,
    ):
        if bases == (SchemaBaseModel,) or not namespace.get(
            "Config", namespace.get("model_config")
        ):
            return super().__new__(mcs, name, bases, namespace)

        config = namespace.get("Config")
        if not config:
            model_config = namespace.get("model_config")
            if model_config:
                config = ModelSchemaConfigAdapter(model_config)

        config_instance = None

        if config:
            config_instance = ModelSchemaConfig(name, config)

        if config_instance and config_instance.model and not config_instance.abstract:
            annotations = namespace.get("__annotations__", {})
            try:
                fields = list(config_instance.model_fields())
            except AttributeError as exc:
                raise ConfigError(
                    f"{exc} (Is `Config.model` a valid Django model class?)"
                ) from exc

            field_values, _seen = {}, set()

            all_fields = {f.name: f for f in fields}
            config_instance.check_invalid_keys(**all_fields)

            for field in chain(fields, annotations.copy()):
                field_name = getattr(
                    field, "name", getattr(field, "related_name", field)
                )

                if (
                    field_name in _seen
                    or (
                        (
                            config_instance.include
                            and field_name not in config_instance.include
                        )
                        or (
                            config_instance.exclude
                            and field_name in config_instance.exclude
                        )
                    )
                    and field_name not in annotations
                ):
                    continue

                _seen.add(field_name)
                if field_name in annotations and field_name in namespace:
                    python_type = annotations.pop(field_name)
                    pydantic_field = namespace[field_name]
                    if (
                        hasattr(pydantic_field, "default_factory")
                        and pydantic_field.default_factory
                    ):
                        pydantic_field = pydantic_field.default_factory()

                elif field_name in annotations:
                    python_type = annotations.pop(field_name)
                    pydantic_field = (
                        None if Optional[python_type] == python_type else Ellipsis
                    )

                else:
                    python_type, pydantic_field = convert_django_field_with_choices(
                        field,
                        registry=config_instance.registry,
                        depth=config_instance.depth,
                        skip_registry=config_instance.skip_registry,
                    )

                    if pydantic_field.default is None:
                        python_type = Optional[python_type]

                    if config_instance.is_field_in_optional(field_name):
                        pydantic_field = ModelSchemaConfig.clone_field(
                            field=pydantic_field, default=None, default_factory=None
                        )
                        python_type = Optional[python_type]

                field_values[field_name] = (python_type, pydantic_field)
            if IS_PYDANTIC_V1:
                cls = super().__new__(mcs, name, bases, namespace)
                return update_class_missing_fields(
                    cls,
                    list(bases),
                    compute_field_annotations(namespace, **field_values),
                )
            return super().__new__(
                mcs, name, bases, compute_field_annotations(namespace, **field_values)
            )
        return super().__new__(mcs, name, bases, namespace)


class SchemaBaseModel(SchemaMixins, BaseModel):
    if not IS_PYDANTIC_V1:

        @classmethod
        def from_orm(cls, obj, **options) -> BaseModel:
            return cls.model_validate(  # type:ignore[attr-defined,no-any-return]
                obj, **options
            )

        def dict(self, *a: Any, **kw: Any) -> Dict[str, Any]:
            # Backward compatibility with pydantic 1.x
            return self.model_dump(*a, **kw)  # type:ignore[attr-defined,no-any-return]

        def json(self, *a, **kw) -> str:
            # Backward compatibility with pydantic 1.x
            return self.model_dump_json(*a, **kw)  # type:ignore[attr-defined,no-any-return]

        @classmethod
        def json_schema(cls) -> Dict[str, Any]:
            return cls.model_json_schema()

        @classmethod
        def schema(cls) -> Dict[str, Any]:
            return cls.json_schema()


class ModelSchema(SchemaBaseModel, metaclass=ModelSchemaMetaclass):
    if IS_PYDANTIC_V1:

        class Config:
            orm_mode = True
            getter_dict = DjangoGetter

    else:
        model_config = {"from_attributes": True}



import typing as t

from django.db.models import Model as DjangoModel

from ..pydanticutils import IS_PYDANTIC_V1
from ..types import DictStrAny
from .getters import DjangoGetter


class BaseMixins:
    def apply_to_model(
        self, model_instance: t.Type[DjangoModel], **kwargs: DictStrAny
    ) -> t.Type[DjangoModel]:
        for attr, value in self.dict(**kwargs).items():  # type:ignore[attr-defined]
            setattr(model_instance, attr, value)
        return model_instance


if not IS_PYDANTIC_V1:
    from pydantic import BaseModel, model_validator
    from pydantic.json_schema import GenerateJsonSchema
    from pydantic_core.core_schema import ValidationInfo

    class BaseMixinsV2(BaseMixins):
        @model_validator(mode="before")
        def _run_root_validator(cls, values: t.Any, info: ValidationInfo) -> t.Any:
            values = DjangoGetter(values, cls, info.context)
            return values

        @classmethod
        def from_orm(cls, obj: t.Any, **options: t.Any) -> BaseModel:
            return cls.model_validate(  # type:ignore[attr-defined,no-any-return]
                obj, **options
            )

        def dict(self, *a: t.Any, **kw: t.Any) -> DictStrAny:
            # Backward compatibility with pydantic 1.x
            return self.model_dump(*a, **kw)  # type:ignore[attr-defined,no-any-return]

        @classmethod
        def json_schema(cls) -> DictStrAny:
            return cls.model_json_schema(  # type:ignore[attr-defined,no-any-return]
                schema_generator=GenerateJsonSchema
            )

        @classmethod
        def schema(cls) -> DictStrAny:
            return cls.json_schema()

    BaseMixins = BaseMixinsV2  # type:ignore[misc]


class SchemaMixins(BaseMixins):
    pass



import typing as t

import pydantic
from django.db.models import Manager, QuerySet
from django.db.models.fields.files import FieldFile

from ..pydanticutils import IS_PYDANTIC_V1

__all__ = [
    "DjangoGetter",
]


class DjangoGetterMixin:
    def _convert_result(self, result: t.Any) -> t.Any:
        if isinstance(result, Manager):
            return list(result.all())

        elif isinstance(result, getattr(QuerySet, "__origin__", QuerySet)):
            return list(result)

        elif isinstance(result, FieldFile):
            if not result:
                return None
            return result.url

        return result


if IS_PYDANTIC_V1:
    from pydantic.utils import GetterDict

    pydantic_version = list(map(int, pydantic.VERSION.split(".")))[:2]
    assert pydantic_version >= [1, 6], "Pydantic 1.6+ required"

    class DjangoGetter(GetterDict, DjangoGetterMixin):
        def get(self, key: t.Any, default: t.Any = None) -> t.Any:
            result = super().get(key, default)
            return self._convert_result(result)

else:

    class DjangoGetter(DjangoGetterMixin):  # type:ignore[no-redef]
        __slots__ = ("_obj", "_schema_cls", "_context")

        def __init__(self, obj: t.Any, schema_cls: t.Any, context: t.Any = None):
            self._obj = obj
            self._schema_cls = schema_cls
            self._context = context

        def __getattr__(self, key: str) -> t.Any:
            # if key.startswith("__pydantic"):
            #     return getattr(self._obj, key)
            if isinstance(self._obj, dict):
                if key not in self._obj:
                    raise AttributeError(key)
                value = self._obj[key]
            else:
                try:
                    value = getattr(self._obj, key)
                except AttributeError as e:
                    raise AttributeError(key) from e

            return self._convert_result(value)



"""Django Schema - Builds Pydantic Schemas from Django Models with default field type validations"""

__version__ = "0.13.6"

from .orm.factory import SchemaFactory
from .orm.model_schema import ModelSchema
from .orm.model_validators import model_validator
from .orm.schema import Schema

__all__ = ["SchemaFactory", "Schema", "ModelSchema", "model_validator"]



import json

import pytest
from pydantic import ValidationError

from ninja_schema import ModelSchema
from ninja_schema.pydanticutils import IS_PYDANTIC_V1
from tests.models import Student, StudentEmail


class TestCustomFields:
    @pytest.mark.skipif(IS_PYDANTIC_V1, reason="requires pydantic == 2.1.x")
    def test_enum_field(self):
        class StudentSchema(ModelSchema):
            model_config = {"model": Student, "include": "__all__"}

        print(json.dumps(StudentSchema.schema(), sort_keys=False, indent=4))
        assert StudentSchema.schema() == {
            "$defs": {
                "SemesterEnum": {
                    "enum": ["1", "2", "3"],
                    "title": "SemesterEnum",
                    "type": "string",
                }
            },
            "properties": {
                "id": {
                    "anyOf": [{"type": "integer"}, {"type": "null"}],
                    "default": None,
                    "description": "",
                    "title": "Id",
                },
                "semester": {
                    "allOf": [{"$ref": "#/$defs/SemesterEnum"}],
                    "default": "1",
                    "description": "",
                    "title": "Semester",
                },
            },
            "title": "StudentSchema",
            "type": "object",
        }
        schema_instance = StudentSchema(semester="1")
        assert str(schema_instance.json()) == '{"id":null,"semester":"1"}'
        with pytest.raises(ValidationError):
            StudentSchema(semester="something")

    @pytest.mark.skipif(IS_PYDANTIC_V1, reason="requires pydantic == 2.1.x")
    def test_email_field(self):
        class StudentEmailSchema(ModelSchema):
            class Config:
                model = StudentEmail
                include = "__all__"

        print(json.dumps(StudentEmailSchema.schema(), sort_keys=False, indent=4))
        assert StudentEmailSchema.schema() == {
            "properties": {
                "id": {
                    "anyOf": [{"type": "integer"}, {"type": "null"}],
                    "default": None,
                    "description": "",
                    "title": "Id",
                },
                "email": {
                    "description": "",
                    "format": "email",
                    "title": "Email",
                    "type": "string",
                },
            },
            "required": ["email"],
            "title": "StudentEmailSchema",
            "type": "object",
        }
        assert (
            str(StudentEmailSchema(email="email@example.com").json())
            == '{"id":null,"email":"email@example.com"}'
        )
        with pytest.raises(ValidationError):
            StudentEmailSchema(email="emailexample.com")



import json
from unittest.mock import Mock

import django
import pytest
from django.db import models
from django.db.models import Manager
from pydantic import ValidationError

from ninja_schema import ModelSchema
from ninja_schema.pydanticutils import IS_PYDANTIC_V1
from tests.models import Week


@pytest.mark.skipif(not IS_PYDANTIC_V1, reason="requires pydantic == 1.6.x")
def test_inheritance():
    class ParentModel(models.Model):
        parent_field = models.CharField()

        class Meta:
            app_label = "tests"

    class ChildModel(ParentModel):
        child_field = models.CharField()

        class Meta:
            app_label = "tests"

    class ChildSchema(ModelSchema):
        class Config:
            model = ChildModel

    print(ChildSchema.schema())

    assert ChildSchema.schema() == {
        "title": "ChildSchema",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "type": "integer"},
            "parent_field": {"title": "Parent Field", "type": "string"},
            "parentmodel_ptr_id": {
                "title": "Parentmodel Ptr",
                "type": "integer",
                "extra": {},
            },
            "child_field": {"title": "Child Field", "type": "string"},
        },
        "required": ["id", "parent_field", "child_field"],
    }


@pytest.mark.skipif(not IS_PYDANTIC_V1, reason="requires pydantic == 1.6.x")
def test_all_fields():
    # test all except relational field

    class AllFields(models.Model):
        bigintegerfield = models.BigIntegerField()
        binaryfield = models.BinaryField()
        booleanfield = models.BooleanField()
        charfield = models.CharField()
        commaseparatedintegerfield = models.CommaSeparatedIntegerField()
        datefield = models.DateField()
        datetimefield = models.DateTimeField()
        decimalfield = models.DecimalField()
        durationfield = models.DurationField()
        emailfield = models.EmailField()
        filefield = models.FileField()
        filepathfield = models.FilePathField()
        floatfield = models.FloatField()
        genericipaddressfield = models.GenericIPAddressField()
        ipaddressfield = models.IPAddressField()
        imagefield = models.ImageField()
        integerfield = models.IntegerField()
        nullbooleanfield = models.NullBooleanField()
        positiveintegerfield = models.PositiveIntegerField()
        positivesmallintegerfield = models.PositiveSmallIntegerField()
        slugfield = models.SlugField()
        smallintegerfield = models.SmallIntegerField()
        textfield = models.TextField()
        timefield = models.TimeField()
        urlfield = models.URLField()
        uuidfield = models.UUIDField()

        class Meta:
            app_label = "tests"

    class AllFieldsSchema(ModelSchema):
        class Config:
            model = AllFields

    # print(SchemaCls.schema())
    assert AllFieldsSchema.schema() == {
        "title": "AllFieldsSchema",
        "type": "object",
        "properties": {
            "id": {"extra": {}, "title": "Id", "type": "integer"},
            "bigintegerfield": {"title": "Bigintegerfield", "type": "integer"},
            "binaryfield": {
                "title": "Binaryfield",
                "type": "string",
                "format": "binary",
            },
            "booleanfield": {"title": "Booleanfield", "type": "boolean"},
            "charfield": {"title": "Charfield", "type": "string"},
            "commaseparatedintegerfield": {
                "title": "Commaseparatedintegerfield",
                "type": "string",
            },
            "datefield": {"title": "Datefield", "type": "string", "format": "date"},
            "datetimefield": {
                "title": "Datetimefield",
                "type": "string",
                "format": "date-time",
            },
            "decimalfield": {"title": "Decimalfield", "type": "number"},
            "durationfield": {
                "title": "Durationfield",
                "type": "number",
                "format": "time-delta",
            },
            "emailfield": {"title": "Emailfield", "format": "email", "type": "string"},
            "filefield": {"title": "Filefield", "type": "string"},
            "filepathfield": {"title": "Filepathfield", "type": "string"},
            "floatfield": {"title": "Floatfield", "type": "number"},
            "genericipaddressfield": {
                "title": "Genericipaddressfield",
                "type": "string",
                "format": "ipvanyaddress",
            },
            "ipaddressfield": {
                "title": "Ipaddressfield",
                "type": "string",
                "format": "ipvanyaddress",
            },
            "imagefield": {"title": "Imagefield", "type": "string"},
            "integerfield": {"title": "Integerfield", "type": "integer"},
            "nullbooleanfield": {"title": "Nullbooleanfield", "type": "boolean"},
            "positiveintegerfield": {
                "title": "Positiveintegerfield",
                "type": "integer",
            },
            "positivesmallintegerfield": {
                "title": "Positivesmallintegerfield",
                "type": "integer",
            },
            "slugfield": {"title": "Slugfield", "type": "string"},
            "smallintegerfield": {"title": "Smallintegerfield", "type": "integer"},
            "textfield": {"title": "Textfield", "type": "string"},
            "timefield": {"title": "Timefield", "type": "string", "format": "time"},
            "urlfield": {
                "title": "Urlfield",
                "type": "string",
                "format": "uri",
                "maxLength": 65536,
                "minLength": 1,
            },
            "uuidfield": {"title": "Uuidfield", "type": "string", "format": "uuid"},
        },
        "required": [
            "bigintegerfield",
            "binaryfield",
            "booleanfield",
            "charfield",
            "commaseparatedintegerfield",
            "datefield",
            "datetimefield",
            "decimalfield",
            "durationfield",
            "emailfield",
            "filefield",
            "filepathfield",
            "floatfield",
            "genericipaddressfield",
            "ipaddressfield",
            "imagefield",
            "integerfield",
            "nullbooleanfield",
            "positiveintegerfield",
            "positivesmallintegerfield",
            "slugfield",
            "smallintegerfield",
            "textfield",
            "timefield",
            "urlfield",
            "uuidfield",
        ],
    }


@pytest.mark.skipif(not IS_PYDANTIC_V1, reason="requires pydantic == 1.6.x")
def test_bigautofield():
    # primary key are optional fields when include = __all__
    class ModelBigAuto(models.Model):
        bigautofiled = models.BigAutoField(primary_key=True)

        class Meta:
            app_label = "tests"

    class ModelBigAutoSchema(ModelSchema):
        class Config:
            model = ModelBigAuto

    print(ModelBigAutoSchema.schema())
    assert ModelBigAutoSchema.schema() == {
        "title": "ModelBigAutoSchema",
        "type": "object",
        "properties": {
            "bigautofiled": {"title": "Bigautofiled", "type": "integer", "extra": {}}
        },
    }


@pytest.mark.skipif(
    django.VERSION < (3, 1), reason="json field introduced in django 3.1"
)
@pytest.mark.skipif(not IS_PYDANTIC_V1, reason="requires pydantic == 1.6.x")
def test_django_31_fields():
    class ModelNewFields(models.Model):
        jsonfield = models.JSONField()
        positivebigintegerfield = models.PositiveBigIntegerField()

        class Meta:
            app_label = "tests"

    class ModelNewFieldsSchema(ModelSchema):
        class Config:
            model = ModelNewFields

    print(ModelNewFieldsSchema.schema())
    assert ModelNewFieldsSchema.schema() == {
        "title": "ModelNewFieldsSchema",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "type": "integer", "extra": {}},
            "jsonfield": {
                "title": "Jsonfield",
                "format": "json-string",
                "type": "string",
            },
            "positivebigintegerfield": {
                "title": "Positivebigintegerfield",
                "type": "integer",
            },
        },
        "required": ["jsonfield", "positivebigintegerfield"],
    }
    with pytest.raises(ValidationError):
        ModelNewFieldsSchema(id=1, jsonfield={"any": "data"}, positivebigintegerfield=1)

    obj = ModelNewFieldsSchema(
        id=1, jsonfield=json.dumps({"any": "data"}), positivebigintegerfield=1
    )
    assert obj.dict() == {
        "id": 1,
        "jsonfield": {"any": "data"},
        "positivebigintegerfield": 1,
    }


@pytest.mark.skipif(not IS_PYDANTIC_V1, reason="requires pydantic == 1.6.x")
def test_relational():
    class Related(models.Model):
        charfield = models.CharField()

        class Meta:
            app_label = "tests"

    class TestModel(models.Model):
        manytomanyfield = models.ManyToManyField(Related)
        onetoonefield = models.OneToOneField(Related, on_delete=models.CASCADE)
        foreignkey = models.ForeignKey(Related, on_delete=models.SET_NULL, null=True)

        class Meta:
            app_label = "tests"

    class TestSchema(ModelSchema):
        class Config:
            model = TestModel

    print(TestSchema.schema())
    assert TestSchema.schema() == {
        "title": "TestSchema",
        "type": "object",
        "properties": {
            "id": {"extra": {}, "title": "Id", "type": "integer"},
            "onetoonefield_id": {"title": "Onetoonefield", "type": "integer"},
            "foreignkey_id": {"title": "Foreignkey", "type": "integer"},
            "manytomanyfield": {
                "title": "Manytomanyfield",
                "type": "array",
                "items": {"type": "integer"},
            },
        },
        "required": ["onetoonefield_id", "manytomanyfield"],
    }


@pytest.mark.skipif(not IS_PYDANTIC_V1, reason="requires pydantic == 1.6.x")
def test_default():
    class MyModel(models.Model):
        default_static = models.CharField(default="hello")
        default_dynamic = models.CharField(default=lambda: "world")

        class Meta:
            app_label = "tests"

    class MyModelSchema(ModelSchema):
        class Config:
            model = MyModel

    assert MyModelSchema.schema() == {
        "title": "MyModelSchema",
        "type": "object",
        "properties": {
            "id": {"title": "Id", "extra": {}, "type": "integer"},
            "default_static": {
                "title": "Default Static",
                "default": "hello",
                "type": "string",
            },
            "default_dynamic": {"title": "Default Dynamic", "type": "string"},
        },
    }


@pytest.mark.skipif(not IS_PYDANTIC_V1, reason="requires pydantic == 1.6.x")
def test_manytomany():
    class Foo(models.Model):
        f = models.CharField()

        class Meta:
            app_label = "tests"

    class Bar(models.Model):
        m2m = models.ManyToManyField(Foo, blank=True)

        class Meta:
            app_label = "tests"

    class BarSchema(ModelSchema):
        class Config:
            model = Bar

    # mocking database data:

    foo = Mock()
    foo.pk = 1
    foo.f = "test"

    m2m = Mock(spec=Manager)
    m2m.all = lambda: [foo]

    bar = Mock()
    bar.id = 1
    bar.m2m = m2m

    data = BarSchema.from_orm(bar).dict()

    assert data == {"id": 1, "m2m": [1]}


@pytest.mark.skipif(not IS_PYDANTIC_V1, reason="requires pydantic == 1.6.x")
def test_manytomany_validation():
    bar = Mock()
    bar.pk = "555555s"

    foo = Mock()
    foo.pk = 1

    class WeekSchema(ModelSchema):
        class Config:
            model = Week

    with pytest.raises(Exception, match="Invalid type"):
        WeekSchema(name="FirstWeek", days=["1", "2"])

    with pytest.raises(Exception, match="Invalid type"):
        WeekSchema(name="FirstWeek", days=[bar, bar])

    schema = WeekSchema(name="FirstWeek", days=[foo, foo])
    assert schema.dict() == {"id": None, "name": "FirstWeek", "days": [1, 1]}



import warnings
from itertools import chain
from types import FunctionType
from typing import Any, Callable, Optional

from ..errors import ConfigError
from ..pydanticutils import IS_PYDANTIC_V1

__all__ = ["model_validator", "ModelValidatorGroup"]

if IS_PYDANTIC_V1:
    from pydantic.class_validators import (
        VALIDATOR_CONFIG_KEY,
        Validator,
        ValidatorGroup,
        _prepare_validator,
    )

    class ModelValidator:
        @classmethod
        def model_validator(
            cls,
            *fields: str,
            pre: bool = False,
            each_item: bool = False,
            always: bool = False,
            check_fields: bool = False,
        ) -> Callable[[Callable], classmethod]:
            """
            Decorate methods on the class indicating that they should be used to validate fields
            :param fields: which field(s) the method should be called on
            :param pre: whether or not this validator should be called before the standard validators (else after)
            :param each_item: for complex objects (sets, lists etc.) whether to validate individual elements rather than the
              whole object
            :param always: whether this method and other validators should be called even if the value is missing
            :param check_fields: whether to check that the fields actually exist on the model
            """
            if not fields:
                raise ConfigError("validator with no fields specified")
            elif isinstance(fields[0], FunctionType):
                raise ConfigError(
                    "validators should be used with fields and keyword arguments, not bare. "  # noqa: Q000
                    "E.g. usage should be `@validator('<field_name>', ...)`"
                )

            def dec(f: Any) -> classmethod:
                f_cls = _prepare_validator(f, True)
                setattr(
                    f_cls,
                    VALIDATOR_CONFIG_KEY,
                    (
                        fields,
                        Validator(
                            func=f_cls.__func__,
                            pre=pre,
                            each_item=each_item,
                            always=always,
                            check_fields=check_fields,
                        ),
                    ),
                )
                return f_cls  # type:ignore[no-any-return]

            return dec

    model_validator = ModelValidator.model_validator

    class ModelValidatorGroup(ValidatorGroup):
        def check_for_unused(self) -> None:
            unused_validators = set(
                chain.from_iterable(
                    (v.func.__name__ for v in self.validators[f])
                    for f in (self.validators.keys() - self.used_validators)
                )
            )
            if unused_validators:
                fn = ", ".join(unused_validators)
                raise ConfigError(
                    f"Validators defined with incorrect fields: {fn} "  # noqa: Q000
                    f"(use check_fields=False if you're inheriting from the model and intended this)"
                )

else:
    from pydantic import field_validator
    from pydantic.functional_validators import FieldValidatorModes

    def model_validator(  # type:ignore[misc]
        __field: str,
        *fields: str,
        mode: FieldValidatorModes = "after",
        check_fields: Optional[bool] = None,
    ) -> Callable[[Any], Any]:
        warnings.warn(
            f"'{model_validator}' is deprecated for pydantic version 2.x.x. "
            f"Use 'field_validator' for the pydantic package instead.",
            category=DeprecationWarning,
            stacklevel=1,
        )
        return field_validator(__field, *fields, mode=mode, check_fields=check_fields)

    class ModelValidatorGroup:  # type:ignore[no-redef]
        pass



import typing
from typing import TYPE_CHECKING, List, Optional, Type, Union, cast

from django.db.models import Model

from ninja_schema.pydanticutils import IS_PYDANTIC_V1

from ..errors import ConfigError
from ..types import DictStrAny
from .schema_registry import SchemaRegister
from .schema_registry import registry as schema_registry

if TYPE_CHECKING:
    from .model_schema import ModelSchema
    from .schema import Schema

__all__ = ["SchemaFactory"]


class SchemaFactory:
    @classmethod
    def get_model_config(cls, **kwargs: DictStrAny) -> Type:
        class Config:
            pass

        for key, value in kwargs.items():
            setattr(Config, key, value)
        return Config

    @classmethod
    def create_schema(
        cls,
        model: Type[Model],
        *,
        registry: SchemaRegister = schema_registry,
        name: str = "",
        depth: int = 0,
        fields: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        skip_registry: bool = False,
        optional_fields: Optional[Union[str, List[str]]] = None,
    ) -> Union[Type["ModelSchema"], Type["Schema"], None]:
        from .model_schema import ModelSchema

        name = name or model.__name__

        if fields and exclude:
            raise ConfigError("Only one of 'include' or 'exclude' should be set.")

        schema = registry.get_model_schema(model)
        if schema and not skip_registry:
            return schema

        model_config_kwargs = {
            "model": model,
            "include": fields,
            "exclude": exclude,
            "skip_registry": skip_registry,
            "depth": depth,
            "registry": registry,
            "optional": optional_fields,
        }
        cls.get_model_config(**model_config_kwargs)  # type: ignore
        new_schema = (
            cls._get_schema_v1(name, model_config_kwargs, ModelSchema)
            if IS_PYDANTIC_V1
            else cls._get_schema_v2(name, model_config_kwargs, ModelSchema)
        )

        new_schema = cast(Type[ModelSchema], new_schema)
        if not skip_registry:
            registry.register_model(model, new_schema)
        return new_schema

    @classmethod
    def _get_schema_v1(
        cls, name: str, model_config_kwargs: typing.Dict, model_type: typing.Type
    ) -> Union[Type["ModelSchema"], Type["Schema"], None]:
        model_config = cls.get_model_config(**model_config_kwargs)

        attrs = {"Config": model_config}

        new_schema = type(name, (model_type,), attrs)
        new_schema = cast(Type["ModelSchema"], new_schema)
        return new_schema

    @classmethod
    def _get_schema_v2(
        cls, name: str, model_config_kwargs: typing.Dict, model_type: typing.Type
    ) -> Union[Type["ModelSchema"], Type["Schema"]]:
        model_config = cls.get_model_config(**model_config_kwargs)

        new_schema_string = f"""class {name}(model_type):
            class Config(model_config):
                pass """

        exec(new_schema_string, locals())
        return locals().get(name)  # type:ignore[return-value]



from pydantic import BaseModel

from ..pydanticutils import IS_PYDANTIC_V1
from .getters import DjangoGetter
from .mixins import SchemaMixins


class Schema(SchemaMixins, BaseModel):
    if IS_PYDANTIC_V1:

        class Config:
            orm_mode = True
            getter_dict = DjangoGetter

    else:
        model_config = {"from_attributes": True}
